/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <ctype.h>
#include "conc.h"

#define CRYPT_BUF 255

typedef struct {
  guint id;
  ConcChannelResponseFunc f;
  gpointer data;
} Command;

/*
  conc_channel_close

    chan - the channel to close

  closes the channel free and frees all resources associated with it.
  the channel close callback is called beforehand.
*/
void
conc_channel_close(ConcChannel *chan)
{
  ConcChannelClosedFunc f;
  GList *l;

  if (!chan->closed_func) return;
  f=chan->closed_func;
  chan->closed_func=NULL;
  f(chan,chan->data);
  g_source_remove(chan->watch_in);
  g_source_remove(chan->watch_hup);
  if (chan->wbuf->len) g_source_remove(chan->watch_out);
  if (chan->rbuf) g_string_free(chan->rbuf,TRUE);
  chan->rbuf=NULL;
  if (chan->wbuf) g_byte_array_free(chan->wbuf,TRUE);
  chan->wbuf=NULL;
  g_io_channel_close(chan->to);
  g_io_channel_unref(chan->to);
  if (chan->from!=chan->to) {
    g_io_channel_close(chan->from);
    g_io_channel_unref(chan->from);
  }
  for (l=chan->commands;l;l=g_list_next(l)) g_free(l->data);
  g_list_free(chan->commands);
  g_free(chan);
}

gboolean
conc_channel_hup(GIOChannel *c, GIOCondition cond, gpointer data)
{
  ConcChannel *chan=(ConcChannel *)data;

  conc_channel_close(chan);
  return FALSE;
}

gboolean
conc_channel_read(GIOChannel *channel, GIOCondition cond, gpointer data)
{
  ConcChannel *chan=(ConcChannel *)data;
  guchar len;
  gint bytes,i,todo;
  gint id;
  gchar id_type='\0';
  GString *line;
  gchar *line_data;
  GList *l;
  Command *c;
  gchar cbuf[CRYPT_BUF+1];
  gchar buf[CRYPT_BUF+1];

  switch (g_io_channel_read(channel,&len,1,&i)) {
    case G_IO_ERROR_NONE:
      break;
    case G_IO_ERROR_AGAIN:
      return TRUE;
    default:
      /*FIXME: report error*/
      break;
  }
  if (!i) {
    conc_channel_close(chan);
    return FALSE;
  }
  bytes=(gint)len;
  for (todo=bytes;todo>0;) {
    switch(g_io_channel_read(channel,&cbuf[bytes-todo],todo,&i)) {
      case G_IO_ERROR_NONE:
        break;
      case G_IO_ERROR_AGAIN:
        i=todo+1;
        break;
      default:
        /*FIXME: report error*/
        break;
    }
    if (i>todo) continue;
    if (!i) {
      conc_channel_close(chan);
      return FALSE;
    }
    todo-=i;
  }
  BF_cfb64_encrypt(cbuf,buf,(long)bytes,&chan->de_key,chan->de_ivec,&chan->de_num,BF_DECRYPT);
  buf[bytes]='\0';
  chan->rbuf=g_string_append(chan->rbuf,buf);
  if (!strstr(chan->rbuf->str,"\n")) {
    if (!bytes) {
      conc_channel_close(chan);
      return FALSE;
    }
    return TRUE;
  }
  while (strstr(chan->rbuf->str,"\n")) {
    if (isdigit(chan->rbuf->str[0])) {
      id=0;
      while (isdigit(chan->rbuf->str[0])) {
        id=id*10+chan->rbuf->str[0]-'0';
        chan->rbuf=g_string_erase(chan->rbuf,0,1);
      }
      id_type=chan->rbuf->str[0];
      chan->rbuf=g_string_erase(chan->rbuf,0,1);
    } else {
      id=-1;
    }
    line=g_string_new("");
    while (chan->rbuf->str[0]!='\n') {
      line=g_string_append_c(line,chan->rbuf->str[0]);
      chan->rbuf=g_string_erase(chan->rbuf,0,1);
    }
    chan->rbuf=g_string_erase(chan->rbuf,0,1); /* drop the \n */
    line_data=line->str; g_string_free(line,FALSE);
    if (id==-1 || id_type=='c') {
      chan->read_func(chan,line_data,id,chan->data);
    } else {
      for (l=chan->commands;l;l=g_list_next(l)) {
        c=(Command *)l->data;
        if (c->id==id) break;
      }
      if (l) {
        c->f(chan,line_data,c->data,data);
        chan->commands=g_list_remove(chan->commands,l->data);
        g_free(c);
      } else {
        /* FIXME: report error - response to unknown command */
      }
    }
  }
  if (!bytes) conc_channel_close(chan);
  return TRUE;
}

/*
  conc_channel_new

  to_fd - the file number to be written to
  from_fd - the file number to be read from (may be the same as to_fd)
  closed_func - the function to call when the channel is about to be closed
  read_func - the function to be called with each line of input
  data - user data
  password - the key for encryption

  returns: a new ConcChannel or NULL if failed
*/
ConcChannel *
conc_channel_new(gint to_fd, gint from_fd,
                 ConcChannelClosedFunc closed_func,
                 ConcChannelReadFunc read_func,
                 gpointer data, const gchar *password
                )
{
  ConcChannel *chan;
  gchar buf[4],cbuf[4];
  gint i;

  if (!read_func) g_error("ConcChannel must have read function");
  if (!closed_func) g_error("ConcChannel must have close function");
  chan=g_new0(ConcChannel,1);
  chan->to=g_io_channel_unix_new(to_fd);
  if (to_fd==from_fd) {
    chan->from=chan->to;
  } else {
    chan->from=g_io_channel_unix_new(from_fd);
  }
  BF_set_key(&chan->en_key,strlen(password),password);
  BF_set_key(&chan->de_key,strlen(password),password);
  BF_cfb64_encrypt("conc",cbuf,4,&chan->en_key,chan->en_ivec,&chan->en_num,BF_ENCRYPT);
  /*FIXME: this could block for ages - need a timeout*/
  g_io_channel_write(chan->to,cbuf,4,&i);
  g_io_channel_read(chan->from,cbuf,4,&i);
  BF_cfb64_encrypt(cbuf,buf,4,&chan->de_key,chan->de_ivec,&chan->de_num,BF_DECRYPT);
  if (i<4||buf[0]!='c'||buf[1]!='o'||buf[2]!='n'||buf[3]!='c') {
    g_io_channel_close(chan->to);
    g_io_channel_unref(chan->to);
    if (chan->from!=chan->to) {
      g_io_channel_close(chan->from);
      g_io_channel_unref(chan->from);
    }
    g_free(chan);
    return NULL;
  }
  chan->rbuf=g_string_new("");
  chan->wbuf=g_byte_array_new();
  chan->closed_func=closed_func;
  chan->read_func=read_func;
  chan->data=data;
  chan->watch_in=g_io_add_watch(chan->from,G_IO_IN,conc_channel_read,chan);
  chan->watch_hup=g_io_add_watch(chan->from,G_IO_HUP,conc_channel_hup,chan);
  return chan;
}

gboolean
conc_channel_send(GIOChannel *channel, GIOCondition cond, gpointer data)
{
  ConcChannel *chan=(ConcChannel *)data;
  GByteArray *newbuf;
  gint i;

  g_io_channel_write(chan->to,chan->wbuf->data,chan->wbuf->len,&i);
  if (i<1) {
    /*FIXME: write fails!*/
    return TRUE;
  }
  if (i==chan->wbuf->len) {
    g_byte_array_free(chan->wbuf,TRUE);
    chan->wbuf=g_byte_array_new();
    return FALSE;
  } else if (i<=(chan->wbuf->len/2)) {
    while (i) {
      chan->wbuf=g_byte_array_remove_index(chan->wbuf,0);
      i--;
    }
  } else {
    newbuf=g_byte_array_new();
    newbuf=g_byte_array_append(newbuf,&chan->wbuf->data[i],chan->wbuf->len-i);
    g_byte_array_free(chan->wbuf,TRUE);
    chan->wbuf=newbuf;
  }
  return TRUE;
}

/*
  conc_channel_write

  chan - the channel to send data to
  line - the line to send (no \n required)
*/
void
conc_channel_write(ConcChannel *chan, const gchar *line)
{
  GString *queue=g_string_new(line);
  gint chunk;
  guchar len;
  gchar cbuf[CRYPT_BUF+1];
  gchar buf[CRYPT_BUF+1];

  queue=g_string_append_c(queue,'\n');
  while (queue->len) {
    chunk=MIN(CRYPT_BUF,queue->len);
    strncpy(buf,queue->str,chunk);
    queue=g_string_erase(queue,0,chunk);
    BF_cfb64_encrypt(buf,cbuf,(long)chunk,&chan->en_key,chan->en_ivec,&chan->en_num,BF_ENCRYPT);
    len=(guchar)chunk;
    if (chan->wbuf->len==0) {
      chan->watch_out=g_io_add_watch(chan->from,G_IO_OUT,conc_channel_send,chan);
    }
    chan->wbuf=g_byte_array_append(chan->wbuf,&len,1);
    chan->wbuf=g_byte_array_append(chan->wbuf,cbuf,chunk);
  }
  g_string_free(queue,TRUE);
}

/*
  conc_channel_respond

  chan - the channel to use
  line - the line to send in response (no \n required)
  id - the id of the command that is being responded to
*/
void
conc_channel_respond(ConcChannel *chan, const gchar *line, gint id)
{
  gchar *s;

  if (id<0) {
    conc_channel_write(chan,line);
    return;
  }
  s=g_strdup_printf("%dr%s",id,line);
  conc_channel_write(chan,s);
}

/*
  conc_channel_command

  chan - the channel to use
  line - the line to send as a command (no \n required)
  response_func - function to call when response received
  command_data - data to supply to the response function
*/
void
conc_channel_command(ConcChannel *chan, const gchar *line,
                     ConcChannelResponseFunc response_func,
                     gpointer command_data)
{
  Command *c;
  gchar *s;

  c=g_new0(Command,1);
  c->id=chan->command_count;
  c->f=response_func;
  c->data=command_data;
  chan->commands=g_list_append(chan->commands,c);
  s=g_strdup_printf("%dc%s",c->id,line);
  conc_channel_write(chan,s);
  g_free(s);
  chan->command_count+=1;
}
