/*
 This file is part of conc.

    conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <openssl/blowfish.h>

typedef struct _ConcChannel ConcChannel;

/*
  ConcChannelClosedFunc

  chan - the channel that is to close
  data - the user data associated with the channel

  this callback is called when the channel is about to be closed
*/
typedef void (*ConcChannelClosedFunc) (ConcChannel *chan, gpointer data);

/*
  ConcChannelReadFunc

  chan - the channel that has been read
  line - the line that was read (\n stripped). must be g_free()'d.
  id - if id>-1 then it is a command id. a response should be made using
       conc_channel_respond() with this id
  data - the user data associated with the channel

  this callback is called when a complete line is read from the channel
  thatis not a respose to a previous command.
*/
typedef void (*ConcChannelReadFunc) (ConcChannel *chan, gchar *line,
                                     gint id, gpointer data);

/*
  ConcChannelResponseFunc

  chan - the channel on which the command was issued
  line - the response (\n stripped). must be g_free()'d
  command_data - the data associated with the command
  data - the user data associated with the channel

  this callback is called when a response is received from an earlier
  command. it is called in preference to ConcChannelReadFunc.
*/
typedef void (*ConcChannelResponseFunc) (ConcChannel *chan, gchar *line,
                                         gpointer command_data,
                                         gpointer data);

/*
  struct _ConcChannel

  all members of this structure should be considered private.
*/
struct _ConcChannel {
  GIOChannel *to;
  GIOChannel *from;
  guint watch_in;
  guint watch_out;
  guint watch_hup;
  guint command_count;
  GList *commands;
  GString *rbuf;
  GByteArray *wbuf; /*may contain nulls*/
  ConcChannelClosedFunc closed_func;
  ConcChannelReadFunc read_func;
  gpointer data;
  BF_KEY en_key;
  BF_KEY de_key;
  gchar en_ivec[8];
  gchar de_ivec[8];
  gint en_num;
  gint de_num;
};

extern ConcChannel *conc_channel_new(gint to_fd, gint from_fd,
                                     ConcChannelClosedFunc closed_func,
                                     ConcChannelReadFunc read_func,
                                     gpointer data, const gchar *password
                                    );

extern void conc_channel_write(ConcChannel *chan, const gchar *line);

extern void conc_channel_respond(ConcChannel *chan, const gchar *line,
                                 gint id);

extern void conc_channel_command(ConcChannel *chan, const gchar *line,
                                 ConcChannelResponseFunc response_func,
                                 gpointer command_data);

extern void conc_channel_close(ConcChannel *chan);
