/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <gnome.h>
#include <gconf/gconf.h>
#include <zvt/zvtterm.h>
#include "conc.h"
#include "conc-ui.h"
#include "tuple.h"

#define DEFAULT_LINES 500

Globals *G;

static void conc_about(GtkWidget *widget, gpointer data);
static void conc_exit(GtkWidget *widget, gpointer data);
static gint conc_console_compare(gconstpointer a, gconstpointer b);
static void conc_show_console(Console *c, GtkWidget *vbox);
static void conc_hide_console(Console *c, gboolean closing);

static GnomeUIInfo file_menu [] = {                     
  GNOMEUIINFO_MENU_EXIT_ITEM(conc_exit,NULL),
  GNOMEUIINFO_END
};                      

static GnomeUIInfo edit_menu [] = {
  GNOMEUIINFO_ITEM(_("Configuration..."),NULL,conc_configure,NULL),
  GNOMEUIINFO_MENU_PREFERENCES_ITEM(conc_preferences,NULL),
  GNOMEUIINFO_END
};
                        
static GnomeUIInfo help_menu [] = {                     
  GNOMEUIINFO_MENU_ABOUT_ITEM(conc_about,NULL),          
  GNOMEUIINFO_END         
};                      
                        
static GnomeUIInfo main_menu [] = {                     
  GNOMEUIINFO_MENU_FILE_TREE(file_menu),          
  GNOMEUIINFO_MENU_EDIT_TREE(edit_menu),          
  GNOMEUIINFO_MENU_HELP_TREE(help_menu),
  GNOMEUIINFO_END         
};  

static int conc_log_close(GnomeDialog *dialog, gpointer data);

static void
make_styles(void)
{
  GdkColor c;

  gdk_color_parse("green",&c);
  G->activity_active=gtk_rc_style_new();
  G->activity_active->bg[GTK_STATE_NORMAL]=c;
  G->activity_active->color_flags[GTK_STATE_NORMAL] |= GTK_RC_BG;
  gdk_color_parse("red",&c);
  G->activity_broken=gtk_rc_style_new();
  G->activity_broken->bg[GTK_STATE_NORMAL]=c;
  G->activity_broken->color_flags[GTK_STATE_NORMAL] |= GTK_RC_BG;
}

static void
ui_flush(void)
{
  while (gtk_events_pending()) gtk_main_iteration();
  gtk_main_iteration_do(FALSE);
  while (gtk_events_pending()) gtk_main_iteration();
  gtk_main_iteration_do(FALSE);
}

static void
conc_exit(GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}

static int
conc_about_close(GnomeDialog *dialog, gpointer data)
{
  GtkWidget *w=(GtkWidget *)data;

  gtk_widget_set_sensitive(w,TRUE);
  return FALSE;
}

static void
conc_about(GtkWidget *widget, gpointer data)
{
  GtkWidget *about;
  const gchar *authors[] = {
          "James Carter <james@jfc.org.uk>",
               NULL
        };

  gtk_widget_set_sensitive(widget,FALSE);
  about=gnome_about_new("Conc",
                        VERSION,
                        "(C) 2002 James Carter/University of York",
                        authors,
                        "Console concentrator - a serial console server",
                        NULL);
  gnome_dialog_set_close(GNOME_DIALOG(about),TRUE);
  gtk_signal_connect(GTK_OBJECT(about), "close",
                     GTK_SIGNAL_FUNC(conc_about_close),
                     widget);
  gtk_widget_show(about);
}

static int                      
delete_event(GtkWidget *widget, GdkEvent *e, gpointer data)                  
{                       
  gtk_main_quit();               
  return FALSE;
}       

static void
send_to_console(gint id,gchar b)
{
  GList *l;
  Console *c;
  int i;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->id!=id) continue;
    if (!c->console_app) break;
    g_io_channel_write(c->to_term,&b,1,&i);
    break;
  }
}

static void
set_console_users(gint id, gchar *num)
{
  GList *l;
  Console *c;
  gchar *s;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->id!=id) continue;
    if (!c->console) break;
    s=g_strdup_printf("%s (%s)",_("Connected Clients"),num);
    gtk_label_set_text(GTK_LABEL(c->console_app_connected),s);
    g_free(s);
    break;
  }
}

static void
activity(gint id,gchar type,gchar *t)
{
  GList *l;
  Console *c;
  gchar *s;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->id!=id) continue;
    if (!c->frame) break;
    s=g_strdup_printf("%s %s",_("Last active:"),t);
    gtk_label_set_text(GTK_LABEL(c->last_active),s);
    g_free(s);
    switch (type) {
      case 'a':
        s="ACTIVE";
        gtk_widget_modify_style(c->activitybox,G->activity_active);
        break;
      case 'i':
        s="INACTIVE";
        gtk_widget_restore_default_style(c->activitybox);
        break;
      default:
        s="DEAD?";
        gtk_widget_modify_style(c->activitybox,G->activity_broken);
        break;
    }
    gtk_label_set_text(GTK_LABEL(c->activity),s);
    break;
  }
  g_free(t);
}

static void
log_read(gint id,gchar *line)
{
  GList *l;
  Console *c;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->id==id) break;
  }
  if (!l || !c->log_app) {
    g_warning("received unexpected log read for id %d",id);
    return;
  }
  gtk_adjustment_set_value(c->log_app_progress_adj,c->log_app_progress_adj->value+1.0);
  gtk_editable_insert_text(GTK_EDITABLE(c->log_app_text),
                           line, strlen(line),
                           &c->log_app_insertpoint);
  gtk_editable_insert_text(GTK_EDITABLE(c->log_app_text),
                           "\n", 1,
                           &c->log_app_insertpoint);
}

static gboolean
get_from_console(GIOChannel *chan, GIOCondition cond, gpointer data)
{
  Console *c=(Console *)data;
  gchar b;
  gchar *s;
  int i=0;

  if (cond==G_IO_IN) g_io_channel_read(chan,&b,1,&i);
  if (!i) {
    close(g_io_channel_unix_get_fd(chan));
    g_io_channel_close(chan);
    g_io_channel_unref(chan);
    return FALSE;
  }
  switch (b) {
    case 0:
      s=g_strdup_printf("z%02x",c->id);
      break;
    case '\n':
      s=g_strdup_printf("T%02x",c->id);
      break;
    default:
      s=g_strdup_printf("t%02x%c",c->id,b);
      break;
  }
  conc_channel_write(G->srv_chan,s);
  g_free(s);
  return TRUE;
}

static void
readremote(ConcChannel *chan, gchar *line, gint id, gpointer data)
{
  GList *l;
  Console *c;
  gchar *s,*o,*n;
  gchar **v;

  switch (line[0]) {
    case 'z':
      send_to_console(getid(&line[1]),0);
      break;
    case 'T':
      send_to_console(getid(&line[1]),'\n');
      break;
    case 't':
      send_to_console(getid(&line[1]),line[3]);
      break;
    case 'l':
      break;
    case 'a':
      activity(getid(&line[1]),line[3],g_strdup(&line[4]));
      break;
    case 'L':
      log_read(getid(&line[1]),&line[3]);
      break;
    case 'U':
      set_console_users(getid(&line[1]),&line[3]);
      break;
    case 'I': /* identify - requires response */
      conc_channel_respond(chan,G->identity,id);
      break;
    case 'c': /* config change */
      switch (line[1]) {
        case 'r': /* remove */
          switch (line[2]) {
            case 's': /* ts */
              s=g_strdup_printf("Server removed %s",&line[3]);
              gnome_app_flash(GNOME_APP(G->app),s);
              g_free(s);
              G->server_list=conc_list_remove(G->server_list,&line[3]);
              if (G->config_app) {
                conc_configure_list_server_changed();
              }
              break;
            case 'c': /* console */
              s=g_strdup_printf("Console removed %s",&line[3]);
              gnome_app_flash(GNOME_APP(G->app),s);
              g_free(s);
              for (l=G->consoles;l;) {
                for (l=G->consoles;l;l=g_list_next(l)) {
                  c=(Console *)l->data;
                  if (!strcmp(c->name,&line[3])) break;
                };
                if (l) {
                  s=g_strdup_printf("%02x%s",c->id,c->name);
                  G->console_list=conc_list_remove(G->console_list,s);
                  conc_remove_console(c);
                  g_free(c->name);
                  g_free(c);
                  conc_save_prefs();
                }
              }
              if (G->config_app) {
                conc_configure_list_console_changed();
              }
              break;
          }
          break;
        case 'c': /* change */
          switch (line[2]) {
            case 'c': /* console */
              v=g_strsplit(&line[3],"\r",0);
              o=v[0]; n=v[1]; g_free(v);
              for (l=G->consoles;l;l=g_list_next(l)) {
                c=(Console *)l->data;
                if (!strcmp(c->name,o)) break;
              };
              if (l) {
               s=g_strdup_printf("%02x%s",c->id,c->name);
               G->console_list=conc_list_remove(G->console_list,s);
               g_free(s);
               s=g_strdup_printf("%02x%s",c->id,n);
               G->console_list=conc_list_add(G->console_list,s,TRUE);
               g_free(s);
               g_free(c->name);
               c->name=g_strdup(n);
               if (c->frame) gtk_frame_set_label(GTK_FRAME(c->frame),n);
               conc_save_prefs();
              }
              if (G->config_app) {
                conc_configure_list_console_changed();
              }
              g_free(o); g_free(n);
              break;
          }
          break;
        case 'n': /* new */
          switch (line[2]) {
            case 's': /* ts */
              G->server_list=conc_list_add(G->server_list,&line[3],FALSE);
              s=g_strdup_printf("New server %s",&line[3]);
              gnome_app_flash(GNOME_APP(G->app),_(s));
              g_free(s);
              if (G->config_app) {
                conc_configure_list_server_changed();
              }
              break;
            case 'c': /* console */
              G->console_list=conc_list_add(G->console_list,&line[3],TRUE);
              s=g_strdup_printf("New console %s",&line[5]);
              gnome_app_flash(GNOME_APP(G->app),_(s));
              g_free(s);
              if (G->config_app) {
                conc_configure_list_console_changed();
              }
              break;
          }
          break;
      }
      break;
    default:
      g_warning("could not parse string from server: %s",line);
      break;
  }
  g_free(line);
}

static gint
gethex(gchar c)
{
  switch (c) {
    case '0': case '1': case '2': case '3': case '4': case '5': case '6':
    case '7': case '8': case '9': return(c-'0');
    case 'a': case 'A': return 10;
    case 'b': case 'B': return 11;
    case 'c': case 'C': return 12;
    case 'd': case 'D': return 13;
    case 'e': case 'E': return 14;
    case 'f': case 'F': return 15;
  }
  return 0;
}

gint
getid(gchar *p)
{
  return gethex(p[0])*16+gethex(p[1]);
}

static void
conc_console_list_users_done(ConcChannel *chan, gchar *line, gpointer cd, gpointer d)
{
  Console *c=(Console *)cd;
  gchar *s;
  GtkWidget *w;
  gint i;

  for (i=1;line[i];i++) if (line[i]=='\r') line[i]='\n';
  s=g_strdup_printf("Users connected to %s:\n%s",c->name,&line[1]);
  w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_INFO,GNOME_STOCK_BUTTON_OK,NULL);
  g_free(s); g_free(line);
  gtk_widget_show_all(w);
  gnome_dialog_run_and_close(GNOME_DIALOG(w));
}

static void
conc_console_list_users(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  gchar *s;

  s=g_strdup_printf("U%02x",c->id);
  conc_channel_command(G->srv_chan,s,conc_console_list_users_done,c);
  g_free(s);
}

static void
conc_console_send_break(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  gchar *s;

  s=g_strdup_printf("b%02x",c->id);
  conc_channel_write(G->srv_chan,s);
  g_free(s);
}

static void
conc_console_close(Console *c)
{
  gchar *s;

  gtk_widget_set_sensitive(c->console_b,TRUE);
  c->console_app=NULL;
  s=g_strdup_printf("ut%02x",c->id);
  conc_channel_write(G->srv_chan,s);
  g_free(s);
  zvt_term_closepty(ZVT_TERM(c->console));
}

static void
conc_console_close_button(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  GtkWidget *w=c->console_app;

  conc_console_close(c);
  gtk_widget_destroy(w);
}

static int
conc_console_close_window(GtkWidget *widget, GdkEvent *e, gpointer data)
{
  Console *c=(Console *)data;

  conc_console_close(c);
  return FALSE;
}

static void
console_process(Console *c,gint infd, gint outfd)
{
  struct termios tio;
  fd_set rd,test;
  char s;

  tcgetattr(0,&tio);
  cfmakeraw(&tio);
  tcsetattr(0,TCSANOW,&tio);
  FD_ZERO(&rd);
  FD_SET(0,&rd);
  FD_SET(infd,&rd);
  for (;;) {
    test=rd;
    if (select(infd+1,&test,NULL,NULL,NULL)==-1) continue;
    if (FD_ISSET(0,&test)) {
      if (read(0,&s,1)!=1) return;
      write(outfd,&s,1);
    }
    if (FD_ISSET(infd,&test)) {
      if (read(infd,&s,1)!=1) return;
      write(1,&s,1);
    }
  }
}

static void
conc_console_show(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  gchar *title,*s;
  gint tofd[2], fromfd[2];
  GtkWidget *y,*x,*w,*z;

  gtk_widget_set_sensitive(GTK_WIDGET(button),FALSE);
  title=g_strdup_printf("%s %s",_("Console:"),c->name);
  c->console_app=gnome_app_new(PACKAGE,title);
  g_free(title);
  gtk_signal_connect(GTK_OBJECT(c->console_app), "delete_event",
                     GTK_SIGNAL_FUNC(conc_console_close_window),
                     c);
  y=gtk_vbox_new(FALSE,GNOME_PAD);
  GTK_WIDGET_UNSET_FLAGS(y,GTK_CAN_FOCUS);
  gnome_app_set_contents(GNOME_APP(c->console_app),y);
  w=gtk_hbox_new(FALSE,GNOME_PAD);
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(y),w,FALSE,FALSE,GNOME_PAD_SMALL);
  z=gtk_label_new(c->name);
  GTK_WIDGET_UNSET_FLAGS(z,GTK_CAN_FOCUS);
  gtk_label_set_justify(GTK_LABEL(z),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(w),z,FALSE,TRUE,GNOME_PAD_SMALL);
  x=gtk_hbutton_box_new();
  GTK_WIDGET_UNSET_FLAGS(x,GTK_CAN_FOCUS);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(x),GNOME_PAD_SMALL);
  gtk_box_pack_end(GTK_BOX(w),x,FALSE,FALSE,GNOME_PAD_SMALL);
  c->console_app_connected=gtk_label_new(_("Connected Clients"));
  w=gtk_button_new();
  gtk_container_add(GTK_CONTAINER(w),c->console_app_connected);
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_box_pack_end(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_signal_connect(GTK_OBJECT(w), "clicked",             
                     GTK_SIGNAL_FUNC(conc_console_list_users),
                     c);
  w=gtk_button_new_with_label(_("Send Break"));
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_box_pack_end(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_signal_connect(GTK_OBJECT(w), "clicked",             
                     GTK_SIGNAL_FUNC(conc_console_send_break),
                     c);
  w=gtk_button_new_with_label(_("Close"));
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_box_pack_end(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_signal_connect(GTK_OBJECT(w), "clicked",             
                     GTK_SIGNAL_FUNC(conc_console_close_button),
                     c);
  x=gtk_hbox_new(FALSE,0);
  GTK_WIDGET_UNSET_FLAGS(x,GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(y),x,FALSE,FALSE,GNOME_PAD_SMALL);
  c->console=zvt_term_new_with_size(80,24);
  zvt_term_set_font_name(ZVT_TERM(c->console),G->console_font);
  zvt_term_set_blink(ZVT_TERM(c->console),FALSE);
  zvt_term_set_bell(ZVT_TERM(c->console),TRUE);
  zvt_term_set_scrollback(ZVT_TERM(c->console),G->console_scrollback);
  zvt_term_set_scroll_on_keystroke(ZVT_TERM (c->console),TRUE);
  zvt_term_set_scroll_on_output(ZVT_TERM(c->console),TRUE);
  zvt_term_set_background (ZVT_TERM(c->console),NULL,0,0);
  zvt_term_set_wordclass (ZVT_TERM(c->console),"-A-Za-z0-9/_:.,?+%=");
  gtk_box_pack_start(GTK_BOX(x),c->console,FALSE,FALSE,0);
  w=gtk_vscrollbar_new(GTK_ADJUSTMENT(ZVT_TERM(c->console)->adjustment));
  GTK_WIDGET_UNSET_FLAGS(w,GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,0);
  gtk_widget_show_all(c->console_app);
  pipe(tofd); pipe(fromfd);
  switch (zvt_term_forkpty(ZVT_TERM(c->console),0)) {
    case -1:
      /* error report */
      break;
    case 0:
      close(tofd[1]); close(fromfd[0]);
      console_process(c,tofd[0],fromfd[1]);
      exit(0);
    default:
      break;
  }
  close(tofd[0]); close(fromfd[1]);
  c->to_term=g_io_channel_unix_new(tofd[1]);
  c->from_term=g_io_channel_unix_new(fromfd[0]);
  s=g_strdup_printf("rt%02x",c->id);
  conc_channel_write(G->srv_chan,s);
  g_free(s);
  g_io_add_watch(c->from_term,G_IO_IN|G_IO_HUP|G_IO_ERR,get_from_console,c);
}

static void
conc_console_reset_done(ConcChannel *chan,gchar *line,gpointer cd, gpointer d)
{
  GtkWidget *w;

  if (line[0]=='o') return;
  w=gnome_message_box_new(&line[1],GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
  gtk_widget_show_all(w);
  gnome_dialog_run_and_close(GNOME_DIALOG(w));
}

static void
conc_console_reset(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  gchar *s;

  s=g_strdup_printf("R%02x",c->id);
  conc_channel_command(G->srv_chan,s,conc_console_reset_done,NULL);
  g_free(s);
}

static int
conc_log_close(GnomeDialog *dialog, gpointer data)
{
  Console *c=(Console *)data;

  gtk_widget_set_sensitive(GTK_WIDGET(c->log_b),TRUE);
  c->log_app=NULL;
  return FALSE;
}

static void
conc_log_error(ConcChannel *chan, gchar *line, gpointer cd, gpointer d)
{
  Console *c=(Console *)cd;
  GtkWidget *w;
  gchar *s;
  
  gtk_text_thaw(GTK_TEXT(c->log_app_text));
  if (line[0]=='o') {
    g_free(line);
    gtk_adjustment_set_value(c->log_app_progress_adj,c->log_app_progress_adj->upper);
    return;
  }
  s=g_strdup_printf("Error reading log:\n%s",&line[1]);
  g_free(line);
  w=gnome_message_box_new(s,GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
  g_free(s);
  gtk_widget_show_all(w);
  gnome_dialog_run_and_close(GNOME_DIALOG(w));
  gnome_dialog_close(GNOME_DIALOG(c->log_app));
}

static void
conc_log_show_ok(GtkButton *button,gpointer data)
{
  GPtrArray *t=data;
  GtkWidget *d=GTK_WIDGET(conc_tuple_get(t,0));
  GtkWidget *lines=GTK_WIDGET(conc_tuple_get(t,1));
  Console *c=(Console *)conc_tuple_get(t,2);
  gchar *s;
  GtkWidget *w;
  gint no_lines;

  s=gtk_editable_get_chars(GTK_EDITABLE(lines),0,-1);
  no_lines=atoi(s); g_free(s);
  gnome_dialog_close(GNOME_DIALOG(d));
  s=g_strdup_printf("%s %s",_("Log for"),c->name);
  c->log_app=gnome_dialog_new(s,GNOME_STOCK_BUTTON_CLOSE,NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(c->log_app),GTK_WINDOW(G->app));
  gtk_window_set_default_size(GTK_WINDOW(c->log_app),400,300);
  gtk_window_set_policy(GTK_WINDOW(c->log_app),FALSE,TRUE,FALSE);
  gtk_signal_connect(GTK_OBJECT(c->log_app), "close",
                     GTK_SIGNAL_FUNC(conc_log_close),
                     c);
  g_free(s);
  gnome_dialog_set_close(GNOME_DIALOG(c->log_app),TRUE);
  c->log_app_text=gtk_text_new(NULL,NULL);
  gtk_text_set_word_wrap(GTK_TEXT(c->log_app_text),FALSE);
  gtk_text_set_line_wrap(GTK_TEXT(c->log_app_text),FALSE);
  w=gtk_scrolled_window_new(GTK_TEXT(c->log_app_text)->hadj,
                            GTK_TEXT(c->log_app_text)->vadj);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(c->log_app)->vbox),w,TRUE,TRUE,GNOME_PAD);
  gtk_container_add(GTK_CONTAINER(w),c->log_app_text);
  gtk_text_set_editable(GTK_TEXT(c->log_app_text),FALSE);
  c->log_app_progress_adj=(GtkAdjustment *)gtk_adjustment_new(0.0,0.0,(gfloat)no_lines,1.0,1.0,1.0);
  w=gtk_progress_bar_new_with_adjustment(c->log_app_progress_adj);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(c->log_app)->vbox),w,FALSE,FALSE,GNOME_PAD);
  gtk_widget_show_all(c->log_app);
  s=g_strdup_printf("L%02x%d",c->id,no_lines);
  c->log_app_insertpoint=0;
  gtk_text_freeze(GTK_TEXT(c->log_app_text));
  conc_channel_command(G->srv_chan,s,conc_log_error,c);
  g_free(s);
}

static void
conc_log_show_cancel(GtkButton *button,gpointer data)
{
  gnome_dialog_close(GNOME_DIALOG(data));
}

static int
conc_log_show_close(GnomeDialog *dialog, gpointer data)
{
  GPtrArray *t=data;
  GtkWidget *b=GTK_WIDGET(conc_tuple_get(t,0));
  Console *c=(Console *)conc_tuple_get(t,1);

  if (!c->log_app) gtk_widget_set_sensitive(GTK_WIDGET(b),TRUE);
  gtk_widget_destroy(GTK_WIDGET(dialog));
  return TRUE;
}

static void
conc_log_show(GtkButton *button,gpointer data)
{
  Console *c=(Console *)data;
  GtkWidget *d,*w,*x,*lines;
  GtkAdjustment *lines_adj;

  gtk_widget_set_sensitive(GTK_WIDGET(button),FALSE);
  d=gnome_dialog_new(_("Log Retrieval"),GNOME_STOCK_BUTTON_OK,GNOME_STOCK_BUTTON_CANCEL,NULL);
  gnome_dialog_set_parent(GNOME_DIALOG(d),GTK_WINDOW(G->app));
  gtk_signal_connect(GTK_OBJECT(d), "close",
                     GTK_SIGNAL_FUNC(conc_log_show_close),
                     conc_tuple(d,button,c,NULL));
  gnome_dialog_close_hides(GNOME_DIALOG(d),TRUE);
  x=gtk_hbox_new(FALSE,GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(d)->vbox),x,FALSE,FALSE,GNOME_PAD);
  w=gtk_label_new(_("No. lines to retrieve:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,0);
  lines_adj=(GtkAdjustment *)gtk_adjustment_new((gfloat)DEFAULT_LINES,100.0,10000.0,100.0,1000.0,10.0);
  lines=gtk_spin_button_new(lines_adj,100.0,0);
  gtk_box_pack_start(GTK_BOX(x),lines,FALSE,FALSE,0);
  w=gtk_label_new("");
  gtk_box_pack_end(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD_SMALL);
  gnome_dialog_button_connect(GNOME_DIALOG(d),1,conc_log_show_cancel,d);
  gnome_dialog_button_connect(GNOME_DIALOG(d),0,conc_log_show_ok,conc_tuple(d,d,lines,c,NULL));
  gtk_widget_show_all(d);
}

static gint
conc_console_compare(gconstpointer a, gconstpointer b)
{
  return g_strcasecmp(((Console *)a)->name,((Console *)b)->name);
}

void
conc_update_consoles(void)
{
  GList *l;
  Console *c;
  gint per_col;
  gint row=0,col=0;
  gint consoles=0;
  GtkWidget *vbox;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->active) consoles++;
    if (c->frame) conc_hide_console(c,FALSE);
  }
  per_col=consoles/G->columns;
  vbox=g_array_index(G->vboxen,GtkWidget *,col);
  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (!c->active) continue;
    conc_show_console(c,vbox);
    row++;
    if (row>per_col) {
      row=0; col++;
      g_assert(col<G->columns);
      vbox=g_array_index(G->vboxen,GtkWidget *,col);
    }
  }
}

void
conc_columns_new(gint newcols)
{
  GList *l;
  Console *c;
  gint i;
  GtkWidget *w;

  for (l=G->consoles;l;l=g_list_next(l)) {
    c=(Console *)l->data;
    if (c->frame) conc_hide_console(c,FALSE);
  }
  for (i=0;i<G->columns;i++){
    gtk_widget_destroy(GTK_WIDGET(g_array_index(G->vboxen,GtkWidget *,0)));
    G->vboxen=g_array_remove_index_fast(G->vboxen,0);
  }
  G->columns=newcols;
  for (i=0;i<G->columns;i++) {
    w=gtk_vbox_new(FALSE,GNOME_PAD);
    gtk_widget_show(w);
    G->vboxen=g_array_append_val(G->vboxen,w);
    gtk_box_pack_start(GTK_BOX(G->x),g_array_index(G->vboxen,GtkWidget *,i),
                       FALSE,FALSE,GNOME_PAD);
  }
  conc_update_consoles();
}

void
conc_add_console(Console *c)
{
  GList *l;
  Console *e;

  for (l=G->consoles;l;l=g_list_next(l)) {
    e=(Console *)l->data;
    if (!strcmp(c->name,e->name)) {
      g_warning("tried to add console %s when it already exists!",c->name);
      return;
    }
  }
  G->consoles=g_list_insert_sorted(G->consoles,c,conc_console_compare);
  if (c->active) conc_update_consoles();
}

void
conc_remove_console(Console *c)
{
  if (c->frame) conc_hide_console(c,TRUE);
  G->consoles=g_list_remove(G->consoles,c);
  if (c->frame) conc_update_consoles();
}

static void
conc_hide_console(Console *c, gboolean closing)
{
  gchar *s;

  if (!c->frame) {
    g_warning("tried to hide console %s when it is not yet displayed",
              c->name);
    return;
  }
  gtk_widget_destroy(c->frame);
  c->frame=NULL;
  s=g_strdup_printf("ua%02x",c->id);
  conc_channel_write(G->srv_chan,s);
  g_free(s);
  if (!closing) return;
  if (c->console_app) conc_console_close(c);
  if (c->log_app) gnome_dialog_close(GNOME_DIALOG(c->log_app));
  c->active=FALSE;
}

static void
conc_show_console(Console *c, GtkWidget *vbox)
{
  GtkWidget *w,*h1,*h2;
  gchar *s;
  GtkStyle *style;
  gint lbearing,rbearing,width,ascent,descent;

  if (c->frame) {
    g_warning("tried to display console %s when it is already displayed",
              c->name);
    return;
  }
  c->frame=gtk_frame_new(c->name);
  gtk_box_pack_start(GTK_BOX(vbox),c->frame,FALSE,FALSE,GNOME_PAD);
  w=gtk_vbox_new(FALSE,GNOME_PAD_SMALL);
  gtk_container_add(GTK_CONTAINER(c->frame),w);
  h1=gtk_hbutton_box_new();
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(h1),GNOME_PAD_SMALL);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(h1),GTK_BUTTONBOX_SPREAD);
  h2=gtk_hbox_new(FALSE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(w),h1,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_box_pack_start(GTK_BOX(w),h2,FALSE,FALSE,GNOME_PAD_SMALL);
  c->console_b=gtk_button_new_with_label(_("View Console"));
  gtk_signal_connect(GTK_OBJECT(c->console_b), "clicked",
                     GTK_SIGNAL_FUNC(conc_console_show),
                     c);
  if (c->console_app) gtk_widget_set_sensitive(GTK_WIDGET(c->console_b),FALSE);
  c->log_b=gtk_button_new_with_label(_("View Log"));
  gtk_signal_connect(GTK_OBJECT(c->log_b), "clicked",
                     GTK_SIGNAL_FUNC(conc_log_show),
                     c);
  if (c->log_app) gtk_widget_set_sensitive(GTK_WIDGET(c->log_b),FALSE);
  c->config_b=gtk_button_new_with_label(_("Reset"));
  gtk_signal_connect(GTK_OBJECT(c->config_b), "clicked",
                     GTK_SIGNAL_FUNC(conc_console_reset),
                     c);
  gtk_box_pack_start(GTK_BOX(h1),c->console_b,FALSE,FALSE,0);
  gtk_box_pack_start(GTK_BOX(h1),c->log_b,FALSE,FALSE,0);
  gtk_box_pack_start(GTK_BOX(h1),c->config_b,FALSE,FALSE,0);
  c->last_active=gtk_label_new(_("Last active: "));
  gtk_label_set_justify(GTK_LABEL(c->last_active),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(h2),c->last_active,FALSE,TRUE,GNOME_PAD_SMALL);
  c->activitybox=gtk_event_box_new();
  gtk_box_pack_end(GTK_BOX(h2),c->activitybox,FALSE,FALSE,GNOME_PAD_SMALL);
  c->activity=gtk_label_new("JFC");
  style=gtk_widget_get_style(c->activity);
  gdk_string_extents(style->font,"INACTIVE",&lbearing,&rbearing,&width,&ascent,&descent);
  gtk_widget_set_usize(c->activity,width+2,ascent+descent+2);
  gtk_container_add(GTK_CONTAINER(c->activitybox),c->activity);
  w=gtk_label_new(_("Activity:"));
  gtk_label_set_justify(GTK_LABEL(w),GTK_JUSTIFY_RIGHT);
  gtk_box_pack_end(GTK_BOX(h2),w,FALSE,FALSE,GNOME_PAD_SMALL);
  gtk_widget_show_all(c->frame);
  s=g_strdup_printf("ra%02x",c->id);
  conc_channel_write(G->srv_chan,s);
  g_free(s);
}

static void
parse_list(ConcChannel *chan, gchar *result, gpointer cd, gpointer d)
{
  gchar **l;
  GList *m;
  Console *c;
  gchar *p,*s;
  int i;

  if (!result) return;
  if (G->server_list) G->server_list=conc_list_free(G->server_list);
  if (G->console_list) G->console_list=conc_list_free(G->console_list);
  s=&result[1];
  for (p=result;*p&&*p!=':';p++);
  if (*p) {
    *p='\0'; p++;
  } else {
    p="";
  }
  if (strlen(s)) {
    l=g_strsplit(s,"/",0);
    for (i=0;l[i];i++) {
      G->server_list=conc_list_add(G->server_list,l[i],FALSE);
    }
    g_strfreev(l);
  }
  if (strlen(p)) {
    l=g_strsplit(p,"/",0);
    for (i=0;l[i];i++) {
      G->console_list=conc_list_add(G->console_list,l[i],TRUE);
      for (m=G->consoles;m;m=g_list_next(m)) {
        c=(Console *)m->data;
        if (strcmp(&l[i][2],c->name)) continue;
        c->id=getid(l[i]);
        if (c->active) break;
        c->active=TRUE;
        if (!c->frame) conc_update_consoles();
        break;
      }
    }
    g_strfreev(l);
  }
  g_free(result);
  return;
}

static int
remote_closed_done(GnomeDialog *dialog, gpointer data)
{
  g_error("lost connection to server");
  return FALSE;
}

static void
remote_closed(ConcChannel *chan, gpointer data)
{
  GtkWidget *w;

  w=gnome_message_box_new(_("Lost connection with server.\nThis program must terminate."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_OK,NULL);
  gtk_signal_connect(GTK_OBJECT(w), "close",
                     GTK_SIGNAL_FUNC(remote_closed_done),
                     NULL);
  gnome_dialog_set_close(GNOME_DIALOG(w),TRUE);
  gtk_widget_show_all(w);
}

static void
get_serverpw(gchar **server,gchar **password)
{
  GtkWidget *d,*x,*w,*srv,*pwd;

  *server=*password=NULL;
  d=gnome_dialog_new(_("Enter server and password"),GNOME_STOCK_BUTTON_OK,NULL);
  x=gtk_hbox_new(FALSE,GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(d)->vbox),x,FALSE,FALSE,GNOME_PAD);
  w=gtk_label_new(_("Server:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD);
  srv=gtk_entry_new();
  if (G->server) gtk_entry_set_text(GTK_ENTRY(srv),G->server);
  gtk_box_pack_start(GTK_BOX(x),srv,TRUE,TRUE,GNOME_PAD);
  w=gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD);
  x=gtk_hbox_new(FALSE,GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(d)->vbox),x,FALSE,FALSE,GNOME_PAD);
  w=gtk_label_new(_("Password:"));
  gtk_box_pack_start(GTK_BOX(x),w,FALSE,FALSE,GNOME_PAD);
  pwd=gtk_entry_new();
  gtk_entry_set_visibility(GTK_ENTRY(pwd),FALSE);
  gtk_box_pack_start(GTK_BOX(x),pwd,TRUE,TRUE,GNOME_PAD);
  gnome_dialog_editable_enters(GNOME_DIALOG(d),GTK_EDITABLE(pwd));
  w=gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(x),w,TRUE,TRUE,GNOME_PAD);
  gnome_dialog_close_hides(GNOME_DIALOG(d),TRUE);
  gnome_dialog_set_close(GNOME_DIALOG(d),TRUE);
  gtk_widget_show_all(d);
  while (gnome_dialog_run_and_close(GNOME_DIALOG(d))!=0) gtk_widget_show_all(d);
  *server=gtk_editable_get_chars(GTK_EDITABLE(srv),0,-1);
  *password=gtk_editable_get_chars(GTK_EDITABLE(pwd),0,-1);
  gtk_widget_destroy(d);
}

int
main(int argc, char*argv[])
{
  GtkWidget *w;
  GError *e;
  gint sock,i;
  struct sockaddr_in sa;
  struct hostent *he;
  gchar *server, *password;

  gnome_init(PACKAGE,VERSION,argc,argv);
  G=g_new0(Globals,1);
  make_styles();
  G->app=gnome_app_new(PACKAGE,_("conc -  Console Concentrator"));
  gtk_signal_connect(GTK_OBJECT(G->app), "delete_event",             
                     GTK_SIGNAL_FUNC(delete_event),
                     NULL);
  gnome_app_create_menus(GNOME_APP(G->app), main_menu);
  G->status=gnome_appbar_new(FALSE,TRUE,GNOME_PREFERENCES_USER);
  gnome_app_set_statusbar(GNOME_APP(G->app),G->status);
  G->x=gtk_hbox_new(FALSE,GNOME_PAD);
  w=gtk_scrolled_window_new(NULL,NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(w),G->x);
  gnome_app_set_contents(GNOME_APP(G->app),w);    
  gtk_widget_show_all(G->app);
  gtk_window_set_default_size(GTK_WINDOW(G->app),353,146);
  gnome_app_flash(GNOME_APP(G->app),_("Loading preferences..."));
  ui_flush();
  if (!gconf_init(argc,argv,&e)) {
    g_error("failed to init gconf: %s",e->message);
    g_error_free(e);
    e=NULL;
  }
  G->gce=gconf_engine_get_default();
  conc_load_prefs();

  G->vboxen=g_array_new(FALSE,FALSE,sizeof (GtkWidget *));
  for (i=0;i<G->columns;i++) {
    w=gtk_vbox_new(FALSE,GNOME_PAD);
    gtk_widget_show(w);
    G->vboxen=g_array_append_val(G->vboxen,w);
    gtk_box_pack_start(GTK_BOX(G->x),g_array_index(G->vboxen,GtkWidget *,i),
                       FALSE,FALSE,GNOME_PAD);
  }
  w=gtk_label_new("");
  gtk_box_pack_start(GTK_BOX(G->x),w,TRUE,TRUE,GNOME_PAD);
  while (!G->srv_chan) {
    get_serverpw(&server,&password);
    gnome_app_flash(GNOME_APP(G->app),_("Contacting console server..."));
    ui_flush();
    sock=socket(AF_INET,SOCK_STREAM,0);
    if (sock==-1) {
      g_error("cannot create socket (%d)",errno);
    }
    if ((he=gethostbyname(server))==NULL) {
      w=gnome_message_box_new(_("Could not find server's address."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_CANCEL,GNOME_STOCK_BUTTON_OK,NULL);
      gtk_widget_show_all(w);
      for (i=0;password[i];i++) password[i]='*';
      g_free(password);
      if (!gnome_dialog_run_and_close(GNOME_DIALOG(w))) return 0;
      continue;
    }
    bzero(&sa,sizeof sa);
    bcopy(he->h_addr_list[0],(char *)&sa.sin_addr,he->h_length);
    sa.sin_family=AF_INET;
    sa.sin_port=htons(PORT);
    if (connect(sock,(struct sockaddr *)&sa,sizeof sa)==-1) {
      i=errno;
      w=gnome_message_box_new(_("Could not connect to console server."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_CANCEL,GNOME_STOCK_BUTTON_OK,NULL);
      gtk_widget_show_all(w);
      for (i=0;password[i];i++) password[i]='*';
      g_free(password);
      if (!gnome_dialog_run_and_close(GNOME_DIALOG(w))) return 0;
      continue;
    }
    G->srv_chan=conc_channel_new(sock,sock,remote_closed,readremote,NULL,password);
    for (i=0;password[i];i++) password[i]='*';
    g_free(password);
    if (!G->srv_chan) {
      w=gnome_message_box_new(_("The password is incorrect."),GNOME_MESSAGE_BOX_ERROR,GNOME_STOCK_BUTTON_CANCEL,GNOME_STOCK_BUTTON_OK,NULL);
      gtk_widget_show_all(w);
      if (!gnome_dialog_run_and_close(GNOME_DIALOG(w))) return 0;
    }
    if (!G->server || strcmp(server,G->server)) {
      if (G->server) g_free(G->server);
      G->server=server;
      conc_save_prefs();
    } else {
      g_free(server);
    }
  }
  G->identity=conc_identity();
  conc_channel_write(G->srv_chan,"rc00");
  conc_channel_command(G->srv_chan,"cR",parse_list,NULL);
  
  gtk_main();
  return(0);
}
