/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <glib.h>
#include "conc.h"

GMainLoop *gml;

ConcChannel *master;

gchar *commandbuf;

typedef struct {
  gchar id[2];
  TermSpec spec;
  speed_t ispeed;
  speed_t ospeed;
  GIOChannel *chan;
  guint watch_id;
} Tty;

GList *ttys=NULL;

Tty *
findtty(gchar *id)
{
  GList *l;
  Tty *t;

  for (l=ttys;l;l=g_list_next(l)) {
    t=(Tty *)l->data;
    if (t->id[0]==id[0]&&t->id[1]==id[1]) break;
  }
  if (!l) t=NULL;
  return t;
}

void
closetty(Tty *t,gboolean free_seg)
{
  g_source_remove(t->watch_id);
  close(g_io_channel_unix_get_fd(t->chan));
  g_io_channel_close(t->chan);
  g_io_channel_unref(t->chan);
  if (!free_seg) return;
  ttys=g_list_remove(ttys,t);
  g_free(t->spec.name);
  g_free(t->spec.server);
  g_free(t->spec.device);
  g_free(t->spec.logfile);
  g_free(t);
}

gboolean
readtty(GIOChannel *chan, GIOCondition cond, gpointer data)
{
  Tty *t=(Tty *)data;
  gint bytes;
  gchar buf[6];
  gchar *s;
  
  g_io_channel_read(chan,&buf[3],1,&bytes);
  if (!bytes) {
    s=g_strdup_printf("econsole %s (%s) closed unexpectedly",t->spec.name,t->spec.device);
    conc_channel_write(master,s);
    g_free(s);
    closetty(t,TRUE);
    return FALSE;
  }
  buf[1]=t->id[0]; buf[2]=t->id[1];
  switch (buf[3]) {
    case 0:
      buf[0]='z';
      buf[3]='\0';
      break;
    case '\n':
      buf[0]='T';
      buf[3]='\0';
      break;
    default:
      buf[0]='t';
      buf[4]='\0';
      break;
  }
  conc_channel_write(master,buf);
  return TRUE;
}

gchar *
new_tty(Tty *t)
{
  gint fd;
  struct termios tio;
  gchar *e;

  fd=open(t->spec.device,O_RDWR);
  if (fd<0) {
    e=g_strdup_printf("ecannot open %s",t->spec.device);
    return e;
  }
  if (tcgetattr(fd,&tio)<0) {
    e=g_strdup_printf("ecannot read attributes on %s",t->spec.device);
    close(fd);
    return e;
  }
  cfmakeraw(&tio);
  cfsetispeed(&tio,t->ispeed);
  cfsetospeed(&tio,t->ospeed);
  tio.c_cflag&=~CSIZE;
  tio.c_cflag|=CLOCAL|CSTOPB|(t->spec.data_bits==8?CS8:CS7);
  if (t->spec.stop_bits==1) tio.c_cflag&=~CSTOPB;
  switch (t->spec.parity) {
    case PARITY_NONE:
      tio.c_cflag&=~PARENB;
      break;
    case PARITY_ODD:
      tio.c_cflag|=PARENB|PARODD;
      break;
    case PARITY_EVEN:
      tio.c_cflag|=PARENB;
      tio.c_cflag&=~PARODD;
      break;
  }
  if (tcsetattr(fd,TCSANOW,&tio)<0) {
    e=g_strdup_printf("ecannot set attributes on %s",t->spec.device);
    close(fd);
    return e;
  }
  t->chan=g_io_channel_unix_new(fd);
  ttys=g_list_append(ttys,t);
  t->watch_id=g_io_add_watch(t->chan,G_IO_IN,readtty,t);
  return NULL;
}

gboolean
itospeed(gint i, speed_t *speed) {
  switch (i) {
    case 0: *speed=B0; break;
    case 50: *speed=B50; break;
    case 75: *speed=B75; break;
    case 110: *speed=B110; break;
    case 134: *speed=B134; break;
    case 150: *speed=B150; break;
    case 200: *speed=B200; break;
    case 300: *speed=B300; break;
    case 600: *speed=B600; break;
    case 1200: *speed=B1200; break;
    case 1800: *speed=B1800; break;
    case 2400: *speed=B2400; break;
    case 4800: *speed=B4800; break;
    case 9600: *speed=B9600; break;
    case 19200: *speed=B19200; break;
    case 38400: *speed=B38400; break;
    case 57600: *speed=B57600; break;
    case 115200: *speed=B115200; break;
    case 230400: *speed=B230400; break;
    default: return FALSE;
  }
  return TRUE;
}

#define E_GARBLED "ecommand garbled"
#define E_BAD "ebad command"
#define E_OK "ok"
void
command(ConcChannel *chan, gchar *line, gint id, gpointer data)
{
  gchar b;
  gchar *p;
  gint i;
  Tty *t;
  TermSpec *spec;

  switch (line[0]) {
    case 'n': /* new terminal to monitor - requires response */
      t=g_new0(Tty,1);
      t->id[0]=line[1];
      t->id[1]=line[2];
      spec=conc_termspec_to(&line[3]);
      t->spec=*spec; g_free(spec);
      itospeed(t->spec.r_rate,&t->ispeed);
      itospeed(t->spec.t_rate,&t->ospeed);
      p=new_tty(t);
      if (p) {
        conc_channel_respond(master,p,id);
        g_free(p);
        g_free(t->spec.name);
        g_free(t->spec.server);
        g_free(t->spec.device);
        g_free(t->spec.logfile);
        g_free(t);
      } else {
        conc_channel_respond(master,E_OK,id);
      }
      break;
    case 'c': /* close terminal - requires response */
      t=findtty(&line[1]);
      if (!t) {
        conc_channel_respond(master,"ebad id",id);
        break;
      }
      closetty(t,TRUE);
      conc_channel_respond(master,E_OK,id);
      break;
    case 'q': /* quit */
      g_main_quit(gml);
      break;
    case 'b': /* send break */
      t=findtty(&line[1]);
      if (!t) {
        conc_channel_write(master,"ebad id");
        break;
      }
      tcsendbreak(g_io_channel_unix_get_fd(t->chan),0);
      break;
    case 't': /* transfer byte */
    case 'z': /* transfer null */
    case 'T': /* transfer \n */
      t=findtty(&line[1]);
      if (!t) {
        conc_channel_write(master,"ebad id");
        break;
      }
      switch (line[0]) {
        case 'z':
          b='\0';
          g_io_channel_write(t->chan,&b,1,&i);
          break;
        case 'T':
          b='\n';
          g_io_channel_write(t->chan,&b,1,&i);
          break;
        case 't':
          g_io_channel_write(t->chan,&line[3],1,&i);
          break;
      }
      break;
    case 'r': /* reset - requires response */
      t=findtty(&line[1]);
      if (!t) {
        conc_channel_write(master,"bad id");
        break;
      }
      closetty(t,FALSE);
      p=new_tty(t);
      if (p) {
        conc_channel_respond(master,p,id);
        g_free(p);
        ttys=g_list_remove(ttys,t);
        g_free(t->spec.name);
        g_free(t->spec.server);
        g_free(t->spec.device);
        g_free(t->spec.logfile);
        g_free(t);
      } else {
        conc_channel_respond(master,E_OK,id);
      }
      break;
    case 'e': /* edit - requires response */
      t=findtty(&line[1]);
      if (!t) {
        conc_channel_write(master,"bad id");
        break;
      }
      closetty(t,FALSE);
      g_free(t->spec.name);
      g_free(t->spec.server);
      g_free(t->spec.device);
      g_free(t->spec.logfile);
      spec=conc_termspec_to(&line[3]);
      t->spec=*spec; g_free(spec);
      itospeed(t->spec.r_rate,&t->ispeed);
      itospeed(t->spec.t_rate,&t->ospeed);
      p=new_tty(t);
      if (p) {
        conc_channel_respond(master,p,id);
        g_free(p);
        ttys=g_list_remove(ttys,t);
        g_free(t->spec.name);
        g_free(t->spec.server);
        g_free(t->spec.device);
        g_free(t->spec.logfile);
        g_free(t);
      } else {
        conc_channel_respond(master,E_OK,id);
      }
      break;
    default: /* command not recognised */
      break;
  }
  g_free(line);
}

void
master_closed(ConcChannel *chan, gpointer data)
{
  g_main_quit(gml);
}

int
main(int argc, char *argv[])
{
  gml=g_main_new(FALSE);
  master=conc_channel_new(1,0,master_closed,command,NULL,CONC_PASSWORD);
  if (master) g_main_run(gml);
  exit(0);
}
