/*
 This file is part of conc.

    Conc is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Conc is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with conc; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <pwd.h>
#include <sys/types.h>
#include <glib.h>
#include "util.h"

#define ANSI(s,i) (s[(i)]==0033&&s[(i)+1]=='[')

TermSpec conc_default_spec={
  8, PARITY_NONE, 1, 9600, 9600, NULL, NULL, NULL
};

gchar time_buf[20]; /* YYYY/MM/DD hh:mm:ss\0 */

/*
  conc_termspec_from

  t - termspec to use

  returns: an ascii respresentation of the TermSpec. this must be g_free()'d
*/
gchar *
conc_termspec_from(const TermSpec *t)
{
  GString *s;
  gchar *p;

  s=g_string_new("");
  s=g_string_append_c(s,t->data_bits==8?'8':'7');
  switch (t->parity) {
    case PARITY_NONE:
      s=g_string_append_c(s,'N'); break;
    case PARITY_EVEN:
      s=g_string_append_c(s,'E'); break;
    case PARITY_ODD:
      s=g_string_append_c(s,'O'); break;
  }
  s=g_string_append_c(s,t->stop_bits==1?'1':'2');
  p=g_strdup_printf("%d\r%d\r",t->t_rate,t->r_rate);
  s=g_string_append(s,p);
  g_free(p);
  if (t->name) s=g_string_append(s,t->name);
  s=g_string_append_c(s,'\r');
  if (t->server) s=g_string_append(s,t->server);
  s=g_string_append_c(s,'\r');
  if (t->device) s=g_string_append(s,t->device);
  s=g_string_append_c(s,'\r');
  if (t->logfile) s=g_string_append(s,t->logfile);
  p=s->str;
  g_string_free(s,FALSE);
  return p;
}

/*
  conc_termspec_to

  str - a string representation of a TermSpec
  returns: a TermSpec built from the string. spec and its pointer contents
           must be g_free()'d.
*/
TermSpec *
conc_termspec_to(const gchar *str)
{
  TermSpec *t=g_new0(TermSpec,1);
  GString *s=g_string_new(str);
  gchar **v;

  t->data_bits=s->str[0]=='8'?8:7;
  switch (s->str[1]) {
    case 'N':
      t->parity=PARITY_NONE; break;
    case 'E':
      t->parity=PARITY_EVEN; break;
    case 'O':
      t->parity=PARITY_ODD; break;
  }
  t->stop_bits=s->str[2]=='1'?1:2;
  s=g_string_erase(s,0,3);
  v=g_strsplit(s->str,"\r",0);
  t->t_rate=atoi(v[0]);
  t->r_rate=atoi(v[1]);
  t->name=g_strdup(v[2]);
  t->server=g_strdup(v[3]);
  t->device=g_strdup(v[4]);
  t->logfile=g_strdup(v[5]);
  g_strfreev(v);
  g_string_free(s,TRUE);
  return t;
}

/*
  get_time

  now - a time to return or NULL for the current time

  returns: a string containing the specified time. must not be freed
           as it is a pointer to a static buffer, the content of which
           will change on every call.
*/
const gchar *
get_time(time_t *now)
{
  struct tm *t;
  time_t n;

  if (now) {
    t=localtime(now);
  } else {
    n=time(NULL);
    t=localtime(&n);
  }
  g_snprintf(time_buf,20,"%04d/%02d/%02d %02d:%02d:%02d",t->tm_year+1900,
             t->tm_mon+1,t->tm_mday,t->tm_hour,t->tm_min,t->tm_sec);
  return time_buf;
}

/*
  conc_identity

  returns: a unique(ish) client identity string which must be g_free()'d
*/
gchar *
conc_identity(void)
{
  struct passwd *pw;
  gchar hostname[32];

  if (gethostname(hostname,32)) g_snprintf(hostname,32,"<nohost>");
  pw=getpwuid(geteuid());
  /* FIXME: should use glib's g_get_user_name() */
  return g_strdup_printf("%s@%s:%d",pw?pw->pw_name:"<nobody>",hostname,getpid());
}

static GString *
striptoalpha(GString *s,gint i)
{
  while (s->str[i]&&!(isalpha(s->str[i]))) s=g_string_erase(s,i,1);
  if (s->str[i]) s=g_string_erase(s,i,1);
  return s;
}

/*
  conc_vtstrip

  s - the string to be processed

  returns: the same string stripped of vt control aharacters any any non
           printing characters
*/
GString *
conc_vtstrip(GString *s)
{
  gint i;

  /* strip vt codes */
  for (i=0;i<s->len;) {
    /* FIXME: incomplete */
    if (ANSI(s->str,i)) {
      s=striptoalpha(s,i);
      continue;
    }
    i++;
  }
  /* strip any remaining control codes */
  for (i=0;i<s->len;) {
    if (!isprint(s->str[i])) {
      s=g_string_erase(s,i,1);
      continue;
    }
    i++;
  }
  return s;
}

static gint
compare_with_id(gconstpointer a, gconstpointer b)
{
  gchar *s1=(gchar *)a;
  gchar *s2=(gchar *)b;

  return g_strcasecmp(&s1[2],&s2[2]);
}

/*
  conc_list_add

  l - the list - or NULL if it does not exist yet
  str - the string to add

  returns: the list
*/
ConcList *
conc_list_add(ConcList *l, const gchar *str, gboolean with_id)
{
  GList *p;
  GCompareFunc func;

  if (!str) {
    g_warning("tried to add null string to list");
    return l;
  }
  for (p=(GList *)l;p;p=g_list_next(p)) {
    if (!strcmp((gchar *)p->data,str)) return l;
  }
  if (with_id) func=compare_with_id; else func=(GCompareFunc)g_strcasecmp;
  (GList *)l=g_list_insert_sorted((GList *)l,(gpointer)g_strdup(str),func);
  return l;
}

/*
  conc_list_remove

  l - the list
  str - the string to remove

  returns: the list
*/
ConcList *
conc_list_remove(ConcList *l, const gchar *str)
{
  GList *p;
  gpointer s;

  for (p=(GList *)l;p;p=g_list_next(p)) {
    if (!strcmp((gchar *)p->data,str)) {
      s=l->data;
      (GList *)l=g_list_remove((GList *)l,p->data);
      g_free(s);
      break;
    }
  }
  return l;
}

/*
  conc_list_length

  returns: the number of strings in the list
*/
gint
conc_list_length(ConcList *l)
{
  return g_list_length((GList *)l);
}

/*
  conc_list

  l - the list
  i - an index

  returns: the string at index i or NULL if out of range
*/
const gchar *
conc_list(ConcList *l,gint i)
{
  GList *p=g_list_nth((GList *)l,i);

  return (const gchar *)p?p->data:NULL;
}

/*
  conc_list_dup

  l - the list to be duplicated

  returns: the new list
*/
ConcList *
conc_list_dup(ConcList *l)
{
  GList *n=NULL;
  GList *p;

  for (p=(GList *)l;p;p=g_list_next(p)) {
    n=g_list_append(n,g_strdup((gchar *)p->data));
  }
  return (ConcList *)n;
}

/*
  conc_list_free

  l - the list to free

  returns: NULL
*/
ConcList *
conc_list_free(ConcList *l)
{
  GList *p;

  for (p=(GList *)l;p;p=g_list_next(p)) {
    g_free(p->data);
  }
  g_list_free((GList *)l);
  return NULL;
}
